/** -*-c-*-
 * DO NOT EDIT
 * This file is automatically generated by tool/generic_erb.rb from
 * template/known_errors.inc.tmpl and defs/known_errors.def.
 */

#ifdef E2BIG
        defined_error("E2BIG", E2BIG)
#else
        undefined_error("E2BIG")
#endif
#ifdef EACCES
        defined_error("EACCES", EACCES)
#else
        undefined_error("EACCES")
#endif
#ifdef EADDRINUSE
        defined_error("EADDRINUSE", EADDRINUSE)
#else
        undefined_error("EADDRINUSE")
#endif
#ifdef EADDRNOTAVAIL
        defined_error("EADDRNOTAVAIL", EADDRNOTAVAIL)
#else
        undefined_error("EADDRNOTAVAIL")
#endif
#ifdef EADV
        defined_error("EADV", EADV)
#else
        undefined_error("EADV")
#endif
#ifdef EAFNOSUPPORT
        defined_error("EAFNOSUPPORT", EAFNOSUPPORT)
#else
        undefined_error("EAFNOSUPPORT")
#endif
#ifdef EAGAIN
        defined_error("EAGAIN", EAGAIN)
#else
        undefined_error("EAGAIN")
#endif
#ifdef EALREADY
        defined_error("EALREADY", EALREADY)
#else
        undefined_error("EALREADY")
#endif
#ifdef EAUTH
        defined_error("EAUTH", EAUTH)
#else
        undefined_error("EAUTH")
#endif
#ifdef EBADARCH
        defined_error("EBADARCH", EBADARCH)
#else
        undefined_error("EBADARCH")
#endif
#ifdef EBADE
        defined_error("EBADE", EBADE)
#else
        undefined_error("EBADE")
#endif
#ifdef EBADEXEC
        defined_error("EBADEXEC", EBADEXEC)
#else
        undefined_error("EBADEXEC")
#endif
#ifdef EBADF
        defined_error("EBADF", EBADF)
#else
        undefined_error("EBADF")
#endif
#ifdef EBADFD
        defined_error("EBADFD", EBADFD)
#else
        undefined_error("EBADFD")
#endif
#ifdef EBADMACHO
        defined_error("EBADMACHO", EBADMACHO)
#else
        undefined_error("EBADMACHO")
#endif
#ifdef EBADMSG
        defined_error("EBADMSG", EBADMSG)
#else
        undefined_error("EBADMSG")
#endif
#ifdef EBADR
        defined_error("EBADR", EBADR)
#else
        undefined_error("EBADR")
#endif
#ifdef EBADRPC
        defined_error("EBADRPC", EBADRPC)
#else
        undefined_error("EBADRPC")
#endif
#ifdef EBADRQC
        defined_error("EBADRQC", EBADRQC)
#else
        undefined_error("EBADRQC")
#endif
#ifdef EBADSLT
        defined_error("EBADSLT", EBADSLT)
#else
        undefined_error("EBADSLT")
#endif
#ifdef EBFONT
        defined_error("EBFONT", EBFONT)
#else
        undefined_error("EBFONT")
#endif
#ifdef EBUSY
        defined_error("EBUSY", EBUSY)
#else
        undefined_error("EBUSY")
#endif
#ifdef ECANCELED
        defined_error("ECANCELED", ECANCELED)
#else
        undefined_error("ECANCELED")
#endif
#ifdef ECAPMODE
        defined_error("ECAPMODE", ECAPMODE)
#else
        undefined_error("ECAPMODE")
#endif
#ifdef ECHILD
        defined_error("ECHILD", ECHILD)
#else
        undefined_error("ECHILD")
#endif
#ifdef ECHRNG
        defined_error("ECHRNG", ECHRNG)
#else
        undefined_error("ECHRNG")
#endif
#ifdef ECOMM
        defined_error("ECOMM", ECOMM)
#else
        undefined_error("ECOMM")
#endif
#ifdef ECONNABORTED
        defined_error("ECONNABORTED", ECONNABORTED)
#else
        undefined_error("ECONNABORTED")
#endif
#ifdef ECONNREFUSED
        defined_error("ECONNREFUSED", ECONNREFUSED)
#else
        undefined_error("ECONNREFUSED")
#endif
#ifdef ECONNRESET
        defined_error("ECONNRESET", ECONNRESET)
#else
        undefined_error("ECONNRESET")
#endif
#ifdef EDEADLK
        defined_error("EDEADLK", EDEADLK)
#else
        undefined_error("EDEADLK")
#endif
#ifdef EDEADLOCK
        defined_error("EDEADLOCK", EDEADLOCK)
#else
        undefined_error("EDEADLOCK")
#endif
#ifdef EDESTADDRREQ
        defined_error("EDESTADDRREQ", EDESTADDRREQ)
#else
        undefined_error("EDESTADDRREQ")
#endif
#ifdef EDEVERR
        defined_error("EDEVERR", EDEVERR)
#else
        undefined_error("EDEVERR")
#endif
#ifdef EDOM
        defined_error("EDOM", EDOM)
#else
        undefined_error("EDOM")
#endif
#ifdef EDOOFUS
        defined_error("EDOOFUS", EDOOFUS)
#else
        undefined_error("EDOOFUS")
#endif
#ifdef EDOTDOT
        defined_error("EDOTDOT", EDOTDOT)
#else
        undefined_error("EDOTDOT")
#endif
#ifdef EDQUOT
        defined_error("EDQUOT", EDQUOT)
#else
        undefined_error("EDQUOT")
#endif
#ifdef EEXIST
        defined_error("EEXIST", EEXIST)
#else
        undefined_error("EEXIST")
#endif
#ifdef EFAULT
        defined_error("EFAULT", EFAULT)
#else
        undefined_error("EFAULT")
#endif
#ifdef EFBIG
        defined_error("EFBIG", EFBIG)
#else
        undefined_error("EFBIG")
#endif
#ifdef EFTYPE
        defined_error("EFTYPE", EFTYPE)
#else
        undefined_error("EFTYPE")
#endif
#ifdef EHOSTDOWN
        defined_error("EHOSTDOWN", EHOSTDOWN)
#else
        undefined_error("EHOSTDOWN")
#endif
#ifdef EHOSTUNREACH
        defined_error("EHOSTUNREACH", EHOSTUNREACH)
#else
        undefined_error("EHOSTUNREACH")
#endif
#ifdef EHWPOISON
        defined_error("EHWPOISON", EHWPOISON)
#else
        undefined_error("EHWPOISON")
#endif
#ifdef EIDRM
        defined_error("EIDRM", EIDRM)
#else
        undefined_error("EIDRM")
#endif
#ifdef EILSEQ
        defined_error("EILSEQ", EILSEQ)
#else
        undefined_error("EILSEQ")
#endif
#ifdef EINPROGRESS
        defined_error("EINPROGRESS", EINPROGRESS)
#else
        undefined_error("EINPROGRESS")
#endif
#ifdef EINTR
        defined_error("EINTR", EINTR)
#else
        undefined_error("EINTR")
#endif
#ifdef EINVAL
        defined_error("EINVAL", EINVAL)
#else
        undefined_error("EINVAL")
#endif
#ifdef EIO
        defined_error("EIO", EIO)
#else
        undefined_error("EIO")
#endif
#ifdef EIPSEC
        defined_error("EIPSEC", EIPSEC)
#else
        undefined_error("EIPSEC")
#endif
#ifdef EISCONN
        defined_error("EISCONN", EISCONN)
#else
        undefined_error("EISCONN")
#endif
#ifdef EISDIR
        defined_error("EISDIR", EISDIR)
#else
        undefined_error("EISDIR")
#endif
#ifdef EISNAM
        defined_error("EISNAM", EISNAM)
#else
        undefined_error("EISNAM")
#endif
#ifdef EKEYEXPIRED
        defined_error("EKEYEXPIRED", EKEYEXPIRED)
#else
        undefined_error("EKEYEXPIRED")
#endif
#ifdef EKEYREJECTED
        defined_error("EKEYREJECTED", EKEYREJECTED)
#else
        undefined_error("EKEYREJECTED")
#endif
#ifdef EKEYREVOKED
        defined_error("EKEYREVOKED", EKEYREVOKED)
#else
        undefined_error("EKEYREVOKED")
#endif
#ifdef EL2HLT
        defined_error("EL2HLT", EL2HLT)
#else
        undefined_error("EL2HLT")
#endif
#ifdef EL2NSYNC
        defined_error("EL2NSYNC", EL2NSYNC)
#else
        undefined_error("EL2NSYNC")
#endif
#ifdef EL3HLT
        defined_error("EL3HLT", EL3HLT)
#else
        undefined_error("EL3HLT")
#endif
#ifdef EL3RST
        defined_error("EL3RST", EL3RST)
#else
        undefined_error("EL3RST")
#endif
#ifdef ELAST
        defined_error("ELAST", ELAST)
#else
        undefined_error("ELAST")
#endif
#ifdef ELIBACC
        defined_error("ELIBACC", ELIBACC)
#else
        undefined_error("ELIBACC")
#endif
#ifdef ELIBBAD
        defined_error("ELIBBAD", ELIBBAD)
#else
        undefined_error("ELIBBAD")
#endif
#ifdef ELIBEXEC
        defined_error("ELIBEXEC", ELIBEXEC)
#else
        undefined_error("ELIBEXEC")
#endif
#ifdef ELIBMAX
        defined_error("ELIBMAX", ELIBMAX)
#else
        undefined_error("ELIBMAX")
#endif
#ifdef ELIBSCN
        defined_error("ELIBSCN", ELIBSCN)
#else
        undefined_error("ELIBSCN")
#endif
#ifdef ELNRNG
        defined_error("ELNRNG", ELNRNG)
#else
        undefined_error("ELNRNG")
#endif
#ifdef ELOOP
        defined_error("ELOOP", ELOOP)
#else
        undefined_error("ELOOP")
#endif
#ifdef EMEDIUMTYPE
        defined_error("EMEDIUMTYPE", EMEDIUMTYPE)
#else
        undefined_error("EMEDIUMTYPE")
#endif
#ifdef EMFILE
        defined_error("EMFILE", EMFILE)
#else
        undefined_error("EMFILE")
#endif
#ifdef EMLINK
        defined_error("EMLINK", EMLINK)
#else
        undefined_error("EMLINK")
#endif
#ifdef EMSGSIZE
        defined_error("EMSGSIZE", EMSGSIZE)
#else
        undefined_error("EMSGSIZE")
#endif
#ifdef EMULTIHOP
        defined_error("EMULTIHOP", EMULTIHOP)
#else
        undefined_error("EMULTIHOP")
#endif
#ifdef ENAMETOOLONG
        defined_error("ENAMETOOLONG", ENAMETOOLONG)
#else
        undefined_error("ENAMETOOLONG")
#endif
#ifdef ENAVAIL
        defined_error("ENAVAIL", ENAVAIL)
#else
        undefined_error("ENAVAIL")
#endif
#ifdef ENEEDAUTH
        defined_error("ENEEDAUTH", ENEEDAUTH)
#else
        undefined_error("ENEEDAUTH")
#endif
#ifdef ENETDOWN
        defined_error("ENETDOWN", ENETDOWN)
#else
        undefined_error("ENETDOWN")
#endif
#ifdef ENETRESET
        defined_error("ENETRESET", ENETRESET)
#else
        undefined_error("ENETRESET")
#endif
#ifdef ENETUNREACH
        defined_error("ENETUNREACH", ENETUNREACH)
#else
        undefined_error("ENETUNREACH")
#endif
#ifdef ENFILE
        defined_error("ENFILE", ENFILE)
#else
        undefined_error("ENFILE")
#endif
#ifdef ENOANO
        defined_error("ENOANO", ENOANO)
#else
        undefined_error("ENOANO")
#endif
#ifdef ENOATTR
        defined_error("ENOATTR", ENOATTR)
#else
        undefined_error("ENOATTR")
#endif
#ifdef ENOBUFS
        defined_error("ENOBUFS", ENOBUFS)
#else
        undefined_error("ENOBUFS")
#endif
#ifdef ENOCSI
        defined_error("ENOCSI", ENOCSI)
#else
        undefined_error("ENOCSI")
#endif
#ifdef ENODATA
        defined_error("ENODATA", ENODATA)
#else
        undefined_error("ENODATA")
#endif
#ifdef ENODEV
        defined_error("ENODEV", ENODEV)
#else
        undefined_error("ENODEV")
#endif
#ifdef ENOENT
        defined_error("ENOENT", ENOENT)
#else
        undefined_error("ENOENT")
#endif
#ifdef ENOEXEC
        defined_error("ENOEXEC", ENOEXEC)
#else
        undefined_error("ENOEXEC")
#endif
#ifdef ENOKEY
        defined_error("ENOKEY", ENOKEY)
#else
        undefined_error("ENOKEY")
#endif
#ifdef ENOLCK
        defined_error("ENOLCK", ENOLCK)
#else
        undefined_error("ENOLCK")
#endif
#ifdef ENOLINK
        defined_error("ENOLINK", ENOLINK)
#else
        undefined_error("ENOLINK")
#endif
#ifdef ENOMEDIUM
        defined_error("ENOMEDIUM", ENOMEDIUM)
#else
        undefined_error("ENOMEDIUM")
#endif
#ifdef ENOMEM
        defined_error("ENOMEM", ENOMEM)
#else
        undefined_error("ENOMEM")
#endif
#ifdef ENOMSG
        defined_error("ENOMSG", ENOMSG)
#else
        undefined_error("ENOMSG")
#endif
#ifdef ENONET
        defined_error("ENONET", ENONET)
#else
        undefined_error("ENONET")
#endif
#ifdef ENOPKG
        defined_error("ENOPKG", ENOPKG)
#else
        undefined_error("ENOPKG")
#endif
#ifdef ENOPOLICY
        defined_error("ENOPOLICY", ENOPOLICY)
#else
        undefined_error("ENOPOLICY")
#endif
#ifdef ENOPROTOOPT
        defined_error("ENOPROTOOPT", ENOPROTOOPT)
#else
        undefined_error("ENOPROTOOPT")
#endif
#ifdef ENOSPC
        defined_error("ENOSPC", ENOSPC)
#else
        undefined_error("ENOSPC")
#endif
#ifdef ENOSR
        defined_error("ENOSR", ENOSR)
#else
        undefined_error("ENOSR")
#endif
#ifdef ENOSTR
        defined_error("ENOSTR", ENOSTR)
#else
        undefined_error("ENOSTR")
#endif
#ifdef ENOSYS
        defined_error("ENOSYS", ENOSYS)
#else
        undefined_error("ENOSYS")
#endif
#ifdef ENOTBLK
        defined_error("ENOTBLK", ENOTBLK)
#else
        undefined_error("ENOTBLK")
#endif
#ifdef ENOTCAPABLE
        defined_error("ENOTCAPABLE", ENOTCAPABLE)
#else
        undefined_error("ENOTCAPABLE")
#endif
#ifdef ENOTCONN
        defined_error("ENOTCONN", ENOTCONN)
#else
        undefined_error("ENOTCONN")
#endif
#ifdef ENOTDIR
        defined_error("ENOTDIR", ENOTDIR)
#else
        undefined_error("ENOTDIR")
#endif
#ifdef ENOTEMPTY
        defined_error("ENOTEMPTY", ENOTEMPTY)
#else
        undefined_error("ENOTEMPTY")
#endif
#ifdef ENOTNAM
        defined_error("ENOTNAM", ENOTNAM)
#else
        undefined_error("ENOTNAM")
#endif
#ifdef ENOTRECOVERABLE
        defined_error("ENOTRECOVERABLE", ENOTRECOVERABLE)
#else
        undefined_error("ENOTRECOVERABLE")
#endif
#ifdef ENOTSOCK
        defined_error("ENOTSOCK", ENOTSOCK)
#else
        undefined_error("ENOTSOCK")
#endif
#ifdef ENOTSUP
        defined_error("ENOTSUP", ENOTSUP)
#else
        undefined_error("ENOTSUP")
#endif
#ifdef ENOTTY
        defined_error("ENOTTY", ENOTTY)
#else
        undefined_error("ENOTTY")
#endif
#ifdef ENOTUNIQ
        defined_error("ENOTUNIQ", ENOTUNIQ)
#else
        undefined_error("ENOTUNIQ")
#endif
#ifdef ENXIO
        defined_error("ENXIO", ENXIO)
#else
        undefined_error("ENXIO")
#endif
#ifdef EOPNOTSUPP
        defined_error("EOPNOTSUPP", EOPNOTSUPP)
#else
        undefined_error("EOPNOTSUPP")
#endif
#ifdef EOVERFLOW
        defined_error("EOVERFLOW", EOVERFLOW)
#else
        undefined_error("EOVERFLOW")
#endif
#ifdef EOWNERDEAD
        defined_error("EOWNERDEAD", EOWNERDEAD)
#else
        undefined_error("EOWNERDEAD")
#endif
#ifdef EPERM
        defined_error("EPERM", EPERM)
#else
        undefined_error("EPERM")
#endif
#ifdef EPFNOSUPPORT
        defined_error("EPFNOSUPPORT", EPFNOSUPPORT)
#else
        undefined_error("EPFNOSUPPORT")
#endif
#ifdef EPIPE
        defined_error("EPIPE", EPIPE)
#else
        undefined_error("EPIPE")
#endif
#ifdef EPROCLIM
        defined_error("EPROCLIM", EPROCLIM)
#else
        undefined_error("EPROCLIM")
#endif
#ifdef EPROCUNAVAIL
        defined_error("EPROCUNAVAIL", EPROCUNAVAIL)
#else
        undefined_error("EPROCUNAVAIL")
#endif
#ifdef EPROGMISMATCH
        defined_error("EPROGMISMATCH", EPROGMISMATCH)
#else
        undefined_error("EPROGMISMATCH")
#endif
#ifdef EPROGUNAVAIL
        defined_error("EPROGUNAVAIL", EPROGUNAVAIL)
#else
        undefined_error("EPROGUNAVAIL")
#endif
#ifdef EPROTO
        defined_error("EPROTO", EPROTO)
#else
        undefined_error("EPROTO")
#endif
#ifdef EPROTONOSUPPORT
        defined_error("EPROTONOSUPPORT", EPROTONOSUPPORT)
#else
        undefined_error("EPROTONOSUPPORT")
#endif
#ifdef EPROTOTYPE
        defined_error("EPROTOTYPE", EPROTOTYPE)
#else
        undefined_error("EPROTOTYPE")
#endif
#ifdef EPWROFF
        defined_error("EPWROFF", EPWROFF)
#else
        undefined_error("EPWROFF")
#endif
#ifdef EQFULL
        defined_error("EQFULL", EQFULL)
#else
        undefined_error("EQFULL")
#endif
#ifdef ERANGE
        defined_error("ERANGE", ERANGE)
#else
        undefined_error("ERANGE")
#endif
#ifdef EREMCHG
        defined_error("EREMCHG", EREMCHG)
#else
        undefined_error("EREMCHG")
#endif
#ifdef EREMOTE
        defined_error("EREMOTE", EREMOTE)
#else
        undefined_error("EREMOTE")
#endif
#ifdef EREMOTEIO
        defined_error("EREMOTEIO", EREMOTEIO)
#else
        undefined_error("EREMOTEIO")
#endif
#ifdef ERESTART
        defined_error("ERESTART", ERESTART)
#else
        undefined_error("ERESTART")
#endif
#ifdef ERFKILL
        defined_error("ERFKILL", ERFKILL)
#else
        undefined_error("ERFKILL")
#endif
#ifdef EROFS
        defined_error("EROFS", EROFS)
#else
        undefined_error("EROFS")
#endif
#ifdef ERPCMISMATCH
        defined_error("ERPCMISMATCH", ERPCMISMATCH)
#else
        undefined_error("ERPCMISMATCH")
#endif
#ifdef ESHLIBVERS
        defined_error("ESHLIBVERS", ESHLIBVERS)
#else
        undefined_error("ESHLIBVERS")
#endif
#ifdef ESHUTDOWN
        defined_error("ESHUTDOWN", ESHUTDOWN)
#else
        undefined_error("ESHUTDOWN")
#endif
#ifdef ESOCKTNOSUPPORT
        defined_error("ESOCKTNOSUPPORT", ESOCKTNOSUPPORT)
#else
        undefined_error("ESOCKTNOSUPPORT")
#endif
#ifdef ESPIPE
        defined_error("ESPIPE", ESPIPE)
#else
        undefined_error("ESPIPE")
#endif
#ifdef ESRCH
        defined_error("ESRCH", ESRCH)
#else
        undefined_error("ESRCH")
#endif
#ifdef ESRMNT
        defined_error("ESRMNT", ESRMNT)
#else
        undefined_error("ESRMNT")
#endif
#ifdef ESTALE
        defined_error("ESTALE", ESTALE)
#else
        undefined_error("ESTALE")
#endif
#ifdef ESTRPIPE
        defined_error("ESTRPIPE", ESTRPIPE)
#else
        undefined_error("ESTRPIPE")
#endif
#ifdef ETIME
        defined_error("ETIME", ETIME)
#else
        undefined_error("ETIME")
#endif
#ifdef ETIMEDOUT
        defined_error("ETIMEDOUT", ETIMEDOUT)
#else
        undefined_error("ETIMEDOUT")
#endif
#ifdef ETOOMANYREFS
        defined_error("ETOOMANYREFS", ETOOMANYREFS)
#else
        undefined_error("ETOOMANYREFS")
#endif
#ifdef ETXTBSY
        defined_error("ETXTBSY", ETXTBSY)
#else
        undefined_error("ETXTBSY")
#endif
#ifdef EUCLEAN
        defined_error("EUCLEAN", EUCLEAN)
#else
        undefined_error("EUCLEAN")
#endif
#ifdef EUNATCH
        defined_error("EUNATCH", EUNATCH)
#else
        undefined_error("EUNATCH")
#endif
#ifdef EUSERS
        defined_error("EUSERS", EUSERS)
#else
        undefined_error("EUSERS")
#endif
#ifdef EWOULDBLOCK
        defined_error("EWOULDBLOCK", EWOULDBLOCK)
#else
        undefined_error("EWOULDBLOCK")
#endif
#ifdef EXDEV
        defined_error("EXDEV", EXDEV)
#else
        undefined_error("EXDEV")
#endif
#ifdef EXFULL
        defined_error("EXFULL", EXFULL)
#else
        undefined_error("EXFULL")
#endif
